/**
  *  Author: Jan Fuchs <fuky@sunstel.asu.cas.cz>
  * Version: 1.0
  *    Date: 01-08-2006
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/limits.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#include <glade/glade.h>
#include <glade/glade-build.h>
#include <GL/freeglut.h>

static struct {
  int width;
  int height;
  int x;
  int y;
} window;

static int naxis1;
static int naxis2;
static int radek = 475;
static int min = 0xFFFF;
static int max = 0;
static int zoom_start = -1;
static int zoom_stop = -1;
static int zoom_start_old = -1;
static int zoom_stop_old = -1;
static int cursor_x = -1;
static int cursor_y = -1;
static int combo_box_count = 0;
static float zoom_y = 1;
static float zoom_y_blue = 1;
static float zoom_y_red = 1;
static float posun = 0;
static float pomer_x;
static float pomer_y;
static char posun_str[100];
static char pocatek_str[100];
static char konec_str[100];
static char fits_dir[PATH_MAX];
static char *p_fits_dir = NULL;
static unsigned short int *data_a;
static unsigned short int *data_b;
static GtkWidget *platno_DA;
static GtkLabel *posun_LB;
static GtkLabel *pocatek_LB;
static GtkLabel *konec_LB;
static GtkSpinButton *radek_SB;
static GtkSpinButton *zoom_y_SB;
static GtkSpinButton *zoom_y_blue_SB;
static GtkSpinButton *zoom_y_red_SB;
static GtkComboBox *a_CB;
static GtkComboBox *b_CB;
static GdkRectangle update_rect;

gboolean quit(GtkWidget *widget, gpointer data);
gboolean fit_a_changed(GtkWidget *widget, GdkEventButton *event);
gboolean fit_b_changed(GtkWidget *widget, GdkEventButton *event);
gboolean update_list_fits(GtkWidget *widget, GdkEventButton *event);
gboolean close_file_chooser_dialog(GtkWidget *button);
gboolean change_dir(GtkWidget *button);
gboolean change_dir_fits(GtkWidget *widget);
gboolean radek_changed(GtkWidget *widget);
gboolean zoom_y_changed(GtkWidget *widget);
gboolean zoom_y_red_changed(GtkWidget *widget);
gboolean zoom_y_blue_changed(GtkWidget *widget);
gboolean window_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer data);
GtkWidget *platno_widget_new(gchar *widget_name, gchar *string1, gchar *string2, gint int1, gint int2);

static void posun2posun_str(void);
static void ovladani(gint keyval);
static void read_data(char *p_filename, unsigned short int *buffer);
static void clear_combo_box(void);
static void read_fits_from_dir(void);
static gboolean realize(GtkWidget *widget, gpointer data);
static gboolean button_press_event(GtkWidget *widget, GdkEventButton *event);
static gboolean configure_event(GtkWidget *widget, GdkEventConfigure *event, gpointer data);
static gboolean platno_DA_expose_event(GtkWidget *widget,GdkEventExpose *event, gpointer data);

gboolean quit(GtkWidget *widget, gpointer data)
{
  gtk_main_quit();

  return TRUE;
}

gboolean fit_a_changed(GtkWidget *widget, GdkEventButton *event)
{
  char *name = NULL;

  if ((name = gtk_combo_box_get_active_text(a_CB)) != NULL) {
    read_data(name, data_a);
    gtk_widget_draw(platno_DA, &update_rect);
  }

  return TRUE;
}

gboolean fit_b_changed(GtkWidget *widget, GdkEventButton *event)
{
  char *name = NULL;

  if ((name = gtk_combo_box_get_active_text(b_CB)) != NULL) {
    read_data(name, data_b);
    gtk_widget_draw(platno_DA, &update_rect);
  }

  return TRUE;
}

gboolean update_list_fits(GtkWidget *widget, GdkEventButton *event)
{
  read_fits_from_dir();

  return TRUE;
}

gboolean close_file_chooser_dialog(GtkWidget *button)
{
  gtk_widget_destroy(gtk_widget_get_toplevel(button));

  return TRUE;
}

gboolean change_dir(GtkWidget *button)
{
  p_fits_dir = gtk_file_chooser_get_filename((GtkFileChooser *)gtk_widget_get_toplevel(button));
  gtk_widget_destroy(gtk_widget_get_toplevel(button));

  return TRUE;
}

gboolean change_dir_fits(GtkWidget *widget)
{
  GladeXML *xml;

  xml = glade_xml_new("./share/ostreni.glade", "file_chooser_dialog", NULL);
  if (!xml) {
    g_warning("Nepodarilo se nacist GUI z XML souboru");
    return FALSE;
  }
 
  glade_xml_signal_autoconnect(xml);

  return TRUE;
}

gboolean radek_changed(GtkWidget *widget)
{
  radek = (int)gtk_spin_button_get_value(radek_SB);
  gtk_widget_draw(platno_DA, &update_rect);

  return TRUE;
}

gboolean zoom_y_changed(GtkWidget *widget)
{
  zoom_y = gtk_spin_button_get_value(zoom_y_SB);
  gtk_widget_draw(platno_DA, &update_rect);

  return TRUE;
}

gboolean zoom_y_blue_changed(GtkWidget *widget)
{
  zoom_y_blue = gtk_spin_button_get_value(zoom_y_blue_SB);
  gtk_widget_draw(platno_DA, &update_rect);

  return TRUE;
}

gboolean zoom_y_red_changed(GtkWidget *widget)
{
  zoom_y_red = gtk_spin_button_get_value(zoom_y_red_SB);
  gtk_widget_draw(platno_DA, &update_rect);

  return TRUE;
}

gboolean window_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
  ovladani(event->keyval);

  return TRUE;
}

GtkWidget *platno_widget_new(gchar *widget_name, gchar *string1, gchar *string2, gint int1, gint int2)
{
  GdkGLConfig *glconfig;

  glconfig = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH | GDK_GL_MODE_DOUBLE);
  if (glconfig == NULL) {
    g_print("*** Cannot find the double-buffered visual.\n");
    g_print("*** Trying single-buffered visual.\n");

    /* Try single-buffered visual */
    glconfig = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH);
    if (glconfig == NULL) {
      g_print("*** No appropriate OpenGL-capable visual found.\n");
      exit(EXIT_FAILURE);
    }
  }

  platno_DA = gtk_drawing_area_new();
  gtk_widget_set_size_request(platno_DA, window.width, window.height);

  gtk_widget_set_gl_capability(platno_DA, glconfig, NULL, TRUE, GDK_GL_RGBA_TYPE);

  gtk_widget_add_events(platno_DA,
			GDK_VISIBILITY_NOTIFY_MASK |
                        GDK_BUTTON_MOTION_MASK |
                        GDK_BUTTON_PRESS_MASK);

  g_signal_connect_after(G_OBJECT(platno_DA), "realize", G_CALLBACK(realize), NULL);
  g_signal_connect(G_OBJECT(platno_DA), "configure_event",    G_CALLBACK(configure_event),        NULL);
  g_signal_connect(G_OBJECT(platno_DA), "expose_event",       G_CALLBACK(platno_DA_expose_event), NULL);
  g_signal_connect(G_OBJECT(platno_DA), "button_press_event", G_CALLBACK(button_press_event),     NULL);

  gtk_widget_show(platno_DA);
  gtk_widget_draw(platno_DA, &update_rect);

  return platno_DA;
}

static void posun2posun_str(void)
{
  sprintf((char *)posun_str, "%0.1f", posun);
}

static void ovladani(gint keyval)
{
  switch (keyval) {
    case GDK_Escape:
      if (zoom_start == -1) {
        posun = 0;
        posun2posun_str();
      }
      zoom_y = 1;
      zoom_y_blue = 1;
      zoom_y_red = 1;
      zoom_start = -1;
      zoom_stop = -1;
      zoom_start_old = -1;
      zoom_stop_old = -1;
      cursor_x = -1;
      cursor_y = -1;
      gtk_spin_button_set_value(zoom_y_SB, zoom_y);
      gtk_spin_button_set_value(zoom_y_blue_SB, zoom_y_blue);
      gtk_spin_button_set_value(zoom_y_red_SB, zoom_y_red);
      gtk_label_set_text(posun_LB, posun_str);
      gtk_widget_draw(platno_DA, &update_rect);
      break;
      
    case GDK_Left:
      posun += 0.1;
      posun2posun_str();
      gtk_label_set_text(posun_LB, posun_str);
      gtk_widget_draw(platno_DA, &update_rect);
      break;

    case GDK_Right:
      posun -= 0.1;
      posun2posun_str();
      gtk_label_set_text(posun_LB, posun_str);
      gtk_widget_draw(platno_DA, &update_rect);
      break;

    case GDK_Up:
      if (zoom_start != -1) {
        ++zoom_start;
        ++zoom_start_old;
        ++zoom_stop;
        ++zoom_stop_old;
        gtk_widget_draw(platno_DA, &update_rect);
      }
      break;

    case GDK_Down:
      if (zoom_start != -1) {
        --zoom_start;
        --zoom_start_old;
        --zoom_stop;
        --zoom_stop_old;
        gtk_widget_draw(platno_DA, &update_rect);
      }
      break;

    default:
      break;
  }
}

static void read_data(char *p_filename, unsigned short int *buffer)
{
  int i = 0;
  int j;
  unsigned short int left;
  unsigned short int right;
  unsigned short int *pixel;
  char header[80];
  char path_file[PATH_MAX];
  FILE *p_fr;

  if (p_fits_dir == NULL) return;
  memset(path_file, 0, 100);
  sprintf(path_file, "%s/%s", p_fits_dir, p_filename);
  if ((p_fr = fopen(path_file, "r")) == NULL) {
    perror("Soubor se nepodarilo otevrit");
    return;
  }

  while (fread(&header, 80, 1, p_fr) != 0) {
    header[3] = '\0';
    if (!strcmp(header, "END")) {
      break;
    }
  }

  while (fread(buffer + i * naxis1, naxis1 * 2, 1, p_fr) > 0) {
    for (j = 0; j < naxis1; j++) {
      pixel = buffer + i * naxis1 + j;
      left  = *pixel >> 8;
      right = *pixel << 8;
      *pixel = left | right;
      *pixel -= 32768;
      if (*pixel > max)  max = *pixel;
      if (*pixel < min)  min = *pixel;
    }
    if (++i >= naxis2) break;
  }

  fclose(p_fr);
}

static void clear_combo_box(void)
{
  int i;

  for (i = 0; i < combo_box_count; i++) {
    gtk_combo_box_remove_text(a_CB, 0);
    gtk_combo_box_remove_text(b_CB, 0);
  }

  combo_box_count = 0;
}

static void read_fits_from_dir(void)
{
  DIR *p_dir;
  struct dirent *dir_entry;

  if (p_fits_dir == NULL) return;
  if ((p_dir = opendir(p_fits_dir)) == NULL) {
    perror("Dir open faiulure");
    exit(EXIT_FAILURE);
  }

  clear_combo_box();
  while ((dir_entry = readdir(p_dir)) != NULL) {
    int len;
    char pripona[5];

    len = strlen(dir_entry->d_name);
    strcpy((char *)pripona, (char *)dir_entry->d_name + len - 4);
    if (!strcmp(pripona, ".fit")) {
      gtk_combo_box_append_text(a_CB, dir_entry->d_name);
      gtk_combo_box_append_text(b_CB, dir_entry->d_name);
      combo_box_count++;
    }
  }
}

static gboolean realize(GtkWidget *widget, gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
  
  if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return FALSE;

  glClearColor(1.0, 1.0, 1.0, 0.0);
  glShadeModel(GL_FLAT);

  gdk_gl_drawable_gl_end(gldrawable);

  return TRUE;
}

static gboolean button_press_event(GtkWidget *widget, GdkEventButton *event)
{
  int x = (int)event->x;
  int y = (int)event->y;

  if ((zoom_start == -1) || (zoom_stop != -1)) {
    zoom_start_old = zoom_start;
    if (zoom_start == -1) {
      zoom_start = x / pomer_x;
    }
    else {
      zoom_start = zoom_start + ((zoom_stop - zoom_start) / ((float)window.width / x));
    }
    zoom_stop = -1;
  }
  else if (zoom_stop == -1) {
    if (zoom_stop_old == -1) {
      zoom_stop = x / pomer_x;
    }
    else {
      zoom_stop = zoom_start_old + ((zoom_stop_old - zoom_start) / ((float)window.width / x));
    }
    zoom_stop_old = zoom_stop;

    if ((zoom_stop != -1) && (zoom_start > zoom_stop)) {
      zoom_start_old = zoom_start;
      zoom_start = zoom_stop;
      zoom_stop = zoom_start_old;
      zoom_start_old = zoom_start;
      zoom_stop_old = zoom_stop;
    }
  }

  cursor_x = x;
  cursor_y = y;

  gtk_widget_draw(platno_DA, &update_rect);

  return TRUE;
}

static gboolean configure_event(GtkWidget *widget, GdkEventConfigure *event, gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
  GLfloat w = widget->allocation.width;
  GLfloat h = widget->allocation.height;

  window.width  = (int)w;
  window.height = (int)h;
  update_rect.width  = window.width;
  update_rect.height = window.height;

  if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return FALSE;

  glViewport(0, 0, (GLsizei)w, (GLsizei)h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, w, 0, h, -1, 1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  gdk_gl_drawable_gl_end(gldrawable);

  return TRUE;
}

static gboolean platno_DA_expose_event(GtkWidget *widget,GdkEventExpose *event, gpointer data)
{
  int start;
  int stop;
  int i;
  GdkGLContext  *glcontext  = gtk_widget_get_gl_context(widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);

  if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return FALSE;

  glViewport(0, 0, (GLsizei)window.width, (GLsizei)window.height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, window.width, 0, window.height, -1, 1);

  glClear(GL_COLOR_BUFFER_BIT);

  //glRasterPos2i(0, 0);
  //glPixelZoom((float)window.width / naxis1, (float)window.height / naxis2);
  //glDrawPixels(naxis1, naxis2, GL_LUMINANCE, GL_UNSIGNED_SHORT, data_a);

  if (zoom_stop != -1) {
    start = zoom_start;
    stop = zoom_stop;
  }
  else if (zoom_start_old != -1) {
    start = zoom_start_old;
    stop = zoom_stop_old;
  }
  else {
    start = 0;
    stop = naxis1;
  }

  sprintf(pocatek_str, "%i", start);
  sprintf(konec_str, "%i", stop);
  gtk_label_set_text(pocatek_LB, pocatek_str);
  gtk_label_set_text(konec_LB, konec_str);

  pomer_x = (float)window.width / (stop - start);
  pomer_y = (float)window.height / ((max / zoom_y) / zoom_y_red);

  glScalef(pomer_x, pomer_y, 0.0);
  glColor3f(1.0, 0.0, 0.0);
  glBegin(GL_LINE_STRIP);
    for (i = 0; i < stop; i++) {
      glVertex2f(i, *(data_a + radek * naxis1 + i + start));
    }
  glEnd();

  glScalef(1.0, 1.0 / pomer_y, 0.0);
  pomer_y = (float)window.height / ((max / zoom_y) / zoom_y_blue);

  glScalef(1.0, pomer_y, 0.0);
  glColor3f(0.0, 0.0, 1.0);
  glBegin(GL_LINE_STRIP);
    for (i = 0; i < stop; i++) {
      glVertex2f((float)i + posun, *(data_b + radek * naxis1 + i + start));
    }
  glEnd();

  if (zoom_stop == -1) {
    glScalef(1.0 / pomer_x, 1.0 / pomer_y, 0.0);
    glColor3f(0.0, 1.0, 0.0);
    glBegin(GL_LINE_STRIP);
      glVertex2i(cursor_x, 0);
      glVertex2i(cursor_x, window.height);
    glEnd();
  }

  if (gdk_gl_drawable_is_double_buffered(gldrawable))
    gdk_gl_drawable_swap_buffers(gldrawable);
  else
    glFlush();

  gdk_gl_drawable_gl_end(gldrawable);

  return TRUE;
}

/**
  * MAIN
 */
int main(int argc, char *argv[])
{
  GladeXML *xml;
  GtkWidget *custom;

  naxis1 = 2030;
  naxis2 = 800;

  if (argc == 2) {
    strcpy(fits_dir, argv[1]);
    p_fits_dir = fits_dir;
  }
  else if (argc == 4) {
    strcpy(fits_dir, argv[1]);
    p_fits_dir = fits_dir;
    naxis1 = atoi(argv[2]);
    naxis2 = atoi(argv[3]);
    if (naxis1 == 2720) radek = 225;
  }

  data_a = (unsigned short int *)malloc(sizeof(unsigned short int) * naxis1 * naxis2);
  data_b = (unsigned short int *)malloc(sizeof(unsigned short int) * naxis1 * naxis2);

  posun_str[0] = '0';
  posun_str[1] = '\0';

  window.width = 1015;
  window.height = 400;
  window.x = 0;
  window.y = 0;
  
  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = window.width;
  update_rect.height = window.height;

  gtk_init(&argc, &argv);
  gtk_gl_init(&argc, &argv);

  xml = glade_xml_new("./share/ostreni.glade", "window", NULL);
  if (!xml) {
    g_warning("Nepodarilo se nacist GUI z XML souboru");
    return 1;
  }
 
  glade_xml_signal_autoconnect(xml);

  custom          = (GtkWidget *)     glade_xml_get_widget(xml, "custom");
  posun_LB        = (GtkLabel *)      glade_xml_get_widget(xml, "posun_LB");
  pocatek_LB      = (GtkLabel *)      glade_xml_get_widget(xml, "pocatek_LB");
  konec_LB        = (GtkLabel *)      glade_xml_get_widget(xml, "konec_LB");
  radek_SB        = (GtkSpinButton *) glade_xml_get_widget(xml, "radek_SB");
  zoom_y_SB       = (GtkSpinButton *) glade_xml_get_widget(xml, "zoom_y_SB");
  zoom_y_blue_SB  = (GtkSpinButton *) glade_xml_get_widget(xml, "zoom_y_blue_SB");
  zoom_y_red_SB   = (GtkSpinButton *) glade_xml_get_widget(xml, "zoom_y_red_SB");
  a_CB            = (GtkComboBox *)   glade_xml_get_widget(xml, "a_CB");
  b_CB            = (GtkComboBox *)   glade_xml_get_widget(xml, "b_CB");

  gtk_spin_button_set_value(radek_SB, radek);
  gtk_spin_button_set_value(zoom_y_SB, zoom_y);
  gtk_spin_button_set_value(zoom_y_blue_SB, zoom_y_blue);
  gtk_spin_button_set_value(zoom_y_red_SB, zoom_y_red);

  gtk_widget_show(custom);
  gtk_main();

  return 0;
}
